import subprocess
POPEN_DEFAULTS = {
    'bufsize'            : 1,
    'executable'         : None,
    'stdin'              : subprocess.PIPE,
    'stdout'             : subprocess.PIPE,
    'stderr'             : subprocess.PIPE,
    'preexec_fn'         : None,
    'close_fds'          : False,
    'shell'              : False,
    'cwd'                : None,
    'env'                : None,
    'universal_newlines' : False,
    'startupinfo'        : None,
    'creationflags'      : 0
}



import numpy as np
from ntai.codex.numpy.defaults import (
    NUMPY_FASTA_ENCODEX,
    INCLUDE_URACIL,
    INCLUDE_REPEAT,
    ENCODING_a,
    ENCODING_c,
    ENCODING_g,
    ENCODING_t,
    ENCODING_u,
    ENCODING_repeat
)

from ntai.codex.numpy.utils import (
    make_encodex_matrix, invert_encodex,
    encode, decode
)

NUMPY_FASTA_ENCODEX_MATRIX = make_encodex_matrix(
    NUMPY_FASTA_ENCODEX,
    ENCODING_a,
    ENCODING_c,
    ENCODING_g,
    ENCODING_t,
    ENCODING_u,
    ENCODING_repeat
)

NUMPY_FASTA_DECODEX = invert_encodex(NUMPY_FASTA_ENCODEX)

class Codex:

    def __init__(
        self,
        include_uracil:bool = INCLUDE_URACIL,
        include_repeat:bool = INCLUDE_REPEAT,
        encodex: dict       = NUMPY_FASTA_ENCODEX,
        decodex: dict       = NUMPY_FASTA_DECODEX,
        encodex_matrix: list = NUMPY_FASTA_ENCODEX_MATRIX
    ):
        '''
        Arguments:
            include_uracil (bool): whether or not uracil should be included in the
                embedding. By default False.

            include_repeat (bool): wehtehr or not repeated masked regions should be
                included in the embedding. By default False.

            encodex (dict): the dictionary converting fasta characters to
                nucleotides.

            decodex (dict): the dictionary converting nucleotide characters to
                fasta characters.
        '''
        self.include_uracil = include_uracil
        self.include_repeat = include_repeat
        self.encodex = encodex
        if decodex is None:
            decodex = invert_encodex(encodex)
        self.decodex = decodex
        if encodex_matrix is None:
            encodex_matrix = make_encodex_matrix(encodex=encodex)
        self.encodex_matrix = encodex_matrix


    def encode(self, sequence):
        return encode(
            sequence,
            encodex = self.encodex,
            encodex_matrix = self.encodex_matrix,
            include_uracil = self.include_uracil,
            include_repeat = self.include_repeat
        )

    def decode(self, encoded):
        return decode(
            encoded,
            encodex = self.encodex,
            encodex_matrix = self.encodex_matrix,
            include_uracil = self.include_uracil,
            include_repeat = self.include_repeat,
            decodex        = self.decodex
        )



import numpy as np
from ntai.codex.numpy.defaults import (
    INCLUDE_URACIL,
    INCLUDE_REPEAT,
    ENCODING_a,
    ENCODING_c,
    ENCODING_g,
    ENCODING_t,
    ENCODING_u,
    ENCODING_repeat
)

def validate_base_encoding(name, encoding):
    if np.cumsum(encoding)[-1] != 1:
        msg = 'base encoding for {} should be one hot!'.format(name)
        raise ValueError(msg)
    if encoding.size != 6:
        msg = 'base encoding for {} should be of shape (6, )!'.format(name)
        raise ValueError(msg)

def make_encodex_matrix(
    encodex:dict,
    a:list       = ENCODING_a,
    c:list       = ENCODING_c,
    g:list       = ENCODING_g,
    t:list       = ENCODING_t,
    u:list       = ENCODING_u,
    repeat:list  = ENCODING_repeat
) -> list:
    _a = np.array(a)
    _c = np.array(c)
    _g = np.array(g)
    _t = np.array(t)
    _u = np.array(u)
    _r = np.array(repeat)
    encodings = [('a', _a),('c', _c),('g', _g,),('t', _t,),('u', _u,),('r', _r)]
    for c, encoding in encodings:
        validate_base_encoding(c, encoding)

    e = np.zeros((6, ), dtype=np.int64) # '-'
    r = _a + g
    y = _c + _t + _u
    k = _g + _t + _u
    m = _a + _c
    s = _c + _g
    w = _a + _t + _u
    b = _c + _g + _t + _u
    d = _a + _g + _t + _u
    h = _a + _c + _t + _u
    v = _a + _c + _g
    n = _a + _c + _g + _t + _u
    E =  e + _r
    A = _a + _r
    C = _c + _r
    G = _g + _r
    T = _t + _r
    U = _u + _r
    R =  r + _r
    Y =  y + _r
    K =  k + _r
    M =  m + _r
    S =  s + _r
    W =  w + _r
    B =  b + _r
    D =  d + _r
    H =  h + _r
    V =  v + _r
    N =  n + _r


    encodex_matrix = np.array([
       # 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,
        e, _a,_c,_g,_t,_u, r, y, k, m, s, w, b, d, h, v, n,
       #17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33
        E,  A, C, G, T, U, R, Y, K, M, S, W, B, D, H, V, N,
    ])
    return encodex_matrix[list(encodex.values())]


def invert_encodex(encodex:dict) -> dict:
    k, v = zip(*encodex.items())
    return dict(zip(v, k))

def find_row(matrix:list, row:list) -> int:
    return np.where(np.all(matrix == row, axis=1))[0][0]

def _char_to_int(char:str, encodex:dict)->int:
    return encodex[char]

def _sequence_to_indices(sequence:str, encodex:dict)->list:
    fn = lambda c: _char_to_int(c, encodex)
    return list(map(fn, sequence))


def lookup_channel_index(
    channel:int,
    encodex:dict,
    encodex_matrix:list
) -> int:
    return np.argmax(encodex_matrix[_char_to_int(channel, encodex)])


def get_channel_indices(
    encodex: dict,
    encodex_matrix: list,
    include_uracil: bool = INCLUDE_URACIL,
    include_repeat: bool = INCLUDE_REPEAT
) -> list:
    # drop uracil or repeat channel as needed
    channels = 'acgt'
    if include_uracil: channels += 'u'
    if include_repeat: channels += '_'
    fn = lambda c: lookup_channel_index(c, encodex, encodex_matrix)
    return list(map(fn, channels))


def encode(
    sequence:str,
    encodex: dict,
    encodex_matrix: list = None,
    include_uracil: bool = INCLUDE_URACIL,
    include_repeat: bool = INCLUDE_REPEAT
):
    # convert from string to integers
    indices = np.array(_sequence_to_indices(sequence, encodex))
    if encodex_matrix is None:
        encodex_matrix = make_encodex_matrix(encodex=encodex)
    # extract the rows from the encoding matrix in order as they appear in seq
    results = encodex_matrix[indices]
    channel_indices = get_channel_indices(encodex, encodex_matrix, include_uracil, include_repeat)
    # filter results
    return results[:, channel_indices].tolist()

def decode(
    encoded:list,
    encodex: dict,
    encodex_matrix: list = None,
    include_uracil: bool = INCLUDE_URACIL,
    include_repeat: bool = INCLUDE_REPEAT,
    decodex: dict = None
):

    if encodex_matrix is None:
        encodex_matrix = make_encodex_matrix(encodex=encodex)
    if decodex is None:
        decodex = invert_encodex(encodex)

    channel_indices = get_channel_indices(encodex, encodex_matrix, include_uracil, include_repeat)
    filtered_matrix = encodex_matrix[:, channel_indices]
    keys = list(map(lambda r: find_row(filtered_matrix, r), encoded))
    return ''.join(list(map(lambda k: decodex[k], keys)))



import os, sys, uuid
from ..terminal.tools import short_arg, long_arg, flag_arg

def getfasta(
    fi:     str,
    bed:    str = '/dev/stdin',
    fo:     str = '/dev/stdout',
    name:   bool = False,
    tab:    bool = False,
    bedOut: bool = False,
    s:      bool = False,
    split:  bool = False
) -> list:
    '''
    Arguments:
        fi (str): input FASTA

        bed (str): <BED/GFF/VCF>

        fo (str): Specify an output file name. By default, output goes to stdout

        name (bool): Use the “name” column in the BED file for the FASTA headers
            in the output FASTA file.

        tab (bool): Report extract sequences in a tab-delimited format instead
            of in FASTA format.

        bedOut (bool): Report extract sequences in a tab-delimited BED format
            instead of in FASTA format.

        s (bool): Force strandedness. If the feature occupies the antisense
            strand, the sequence will be reverse complemented. Default: strand
            information is ignored.

        split (bool): Given BED12 input, extract and concatenate the sequences
            from the BED “blocks” (e.g., exons)
    Returns:
        command (list): a list of strings containing the bedtools getfasta
            command
    '''
    if name: name = uuid.uuid4()
    if tab: tab = uuid.uuid4()
    if bedOut: bedOut = uuid.uuid4()
    if s: s = uuid.uuid4()
    if split: split = uuid.uuid4()
    _fi = fi
    del fi
    fi = _fi
    command = [
        'bedtools',
        'getfasta',
        *short_arg(fi, 'fi'), *short_arg(bed, 'bed'), *short_arg(fo, 'fo'),
        flag_arg(name, 'name', short=True), flag_arg(tab, 'tab', short=True),
        flag_arg(bedOut, 'bedOut', short=True), flag_arg(s, 's', short=True),
        flag_arg(split, 'split', short=True)
    ]
    return command



import os, sys, subprocess, uuid
from .utils import variable_name, error_message
from .popen import POPEN_DEFAULTS


def line_arg(arg, name=None):
    if name is None: name = variable_name(arg)
    return '{} {}'.format(name, arg)

def short_arg(arg, name=None):
    '''Converts <arg>, <name> into `['-<arg>', 'name']`'''
    return '-{}'.format(line_arg(arg, name)).split(' ')

def long_arg(arg, name=None):
    '''Converts <arg>, <name> into `['--<arg>', 'name']`'''
    return '--{}'.format(line_arg(arg, name)).split(' ')

def flag_arg(flag, name=None, short=True):
    '''
    Converts <flag>, <short> into `'-<flag>'` or `'--<flag>'` depending on <short>
    '''
    if name is None: name = variable_name(flag)
    dashes = '-' if short else '--'
    return '{}{}'.format(dashes, name) if flag else ''

def clean_command(command:list):
    cleaned = [
        sub
        for part in command
        for sub in part.split(' ')
        if sub != ''
    ]
    return cleaned



def process(command:list, stdin:str, popen_options={}):
    '''
    Arguments:
        command (list): a list of strings indicating the command and its
            arguments to spawn as a subprocess.

        stdin (str): passed as stdin to the subprocess. Assumed to be utf-8
            encoded.

        popen_options (dict): used to configure the subprocess.Popen command

    Returns:
        stdout, stderr
    '''
    command = clean_command(command)
    popen_config = POPEN_DEFAULTS.copy()
    popen_config.update(popen_options)
    try:
        pid = subprocess.Popen(args = command, **popen_config)
        stdout_data, stderr_data = pid.communicate(stdin)
    except OSError as err:
        error_message(command, err)
        sys.exit(1)
    if pid.returncode != 0:
        error_message(command, 'pid code {}'.format(pid.returncode), stdout_data, stderr_data)
    return stdout_data, stderr_data



import os

'''-----------------------------------------------------------------------------
DIRECTORIES AND FILES
-----------------------------------------------------------------------------'''
MODULE_DIR      = os.path.dirname(os.path.realpath(__file__))
PROJECT_DIR     = os.path.join(MODULE_DIR, '..')
DATA_DIR        = os.path.join(PROJECT_DIR, 'data')
SOURCE_DIR      = os.path.join(DATA_DIR, 'source')
DSOURCE_DIR     = os.path.join(DATA_DIR, 'dsource')
DERIVED_DIR     = os.path.join(DATA_DIR, 'derived')
HG38_FILE       = os.path.join(SOURCE_DIR, 'hg38.fa')
JUYPTER_DIR     = os.path.join(PROJECT_DIR, 'juypter')
EXPERIMENTS_DIR = os.path.join(PROJECT_DIR, 'experiments')



from .codex import Codex, NPCodex
from .bedtools import bedtools
from .fetch import fetch_files
from .labeler import Labeler
# from .ranges import LabeledRange, LabeledRanges

name = 'ntai'
version = '0.0.13'
description = 'Nucleotide A.I.'



import urllib.parse, requests, os, gzip, shutil


HG_TABLES_URL = 'http://genome.ucsc.edu/cgi-bin/hgTables'

HG_DOWNLOAD_DIRECTORY_URL = 'http://hgdownload.soe.ucsc.edu'

HG38_DOWNLOAD_DIRECTORY_URL = '{}/goldenPath/hg38/bigZips'.format(HG_DOWNLOAD_DIRECTORY_URL)

HG38_FASTA_URL = '{}/hg38.fa.gz'.format(HG38_DOWNLOAD_DIRECTORY_URL)
HG38_CHROM_SIZES_URL = '{}/hg38.chrom.sizes'.format(HG38_DOWNLOAD_DIRECTORY_URL)



DEFAULT_PARAMS = {
    'ignoreCookie': 1,
    'hgta_track': 'wgEncodeGencodeCompV28',
    'boolshad.hgta_printCustomTrackHeaders':0,
    'hgta_ctName':'tb_wgEncodeGencodeCompV28',
    'hgta_ctDesc': 'table+browser+query+on+wgEncodeGencodeCompV28',
    'hgta_ctVis': 'pack',
    'hgta_ctUrl': None,
    'fbUpBases': 200,
    'fbQual': 'exon',
    'fbExonBases': 0,
    'fbIntronBases': 0,
    'fbDownBases': 200,
    'hgta_doGetBed': 'get+BED'
}


def fname(file:str, directory:str=None):
    '''
    Notes:
        1. if directory is None and file is just the basename, directory is set
            to '~/Downloads'
        2. if file is the full path and directory is set to none, directory is
            ignored.
        3. if file is partial path and directory is not none, directory is
            prepended to the file.
    Arguments:
        file (str): the file's name, either basename or full path.
        directory (str): the directory of the file. Default is None.
    Returns:
        fullpath (str): full path of the file given the directory

    '''
    # user passes just the filename. Then defaults to download directory
    if directory is None and os.path.basename(file) == file:
        directory = '~/Downloads'

    if directory is None:
        directory = ''

    name = os.path.expanduser(os.path.join(directory, file))
    return name




def fetch_exons(
    file:str='encode_gencode_comp_v28_exons.bed',
    directory:str=None
):
    params = {'fbQual': 'exon'}
    request = requests.get(HG_TABLES_URL, params={**DEFAULT_PARAMS, **params})
    if request.status_code == 200:
        with open(fname(file, directory), 'w') as f:
            f.write(request.text)

def fetch_introns(
    file:str='encode_gencode_comp_v28_introns.bed',
    directory:str=None
):
    params = {'fbQual': 'intron'}
    request = requests.get(HG_TABLES_URL, params={**DEFAULT_PARAMS, **params})
    if request.status_code == 200:
        with open(fname(file, directory), 'w') as f:
            f.write(request.text)

def fetch_hg38_chrom_sizes(
    file:str='hg38.chrom.sizes',
    directory:str=None
):
    request = requests.get(HG38_CHROM_SIZES_URL)
    if request.status_code == 200:
        with open(fname(file, directory), 'w') as f:
            f.write(request.text)

def fetch_hg38(
    file:str="hg38.fa.gz",
    directory:str=None
):
    request = requests.get(HG38_FASTA_URL)
    if request.status_code == 200:
        with open(fname(file, directory), 'wb') as f:
            # f.write(request.raw.read())
            f.write(request.content)

def decompress(path:str, odir:str=None):
    '''
    Decompresses the .gz file specified by path. Decompressed file keeps the
    same basename e.g. my_file.txt.gz --> my_file.txt

    Arguments:
        path (str): full path to a .gz compressed file
        odir (str): optional. A new output directory for the file. By default None.

    Returns:
        None.
    '''
    ofile = os.path.splitext(os.path.expanduser(path))[0]
    if odir is not None:
        ofile = os.path.join(odir, os.path.basename(ofile))
    with gzip.open(path, 'rb') as f_in:
        with open(ofile, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)


def fetch_files(directory:str):
    fetch_exons(directory=directory)
    fetch_introns(directory=directory)
    fetch_hg38_chrom_sizes(directory=directory)
    fetch_hg38(directory=directory)
    decompress(os.path.join(directory, 'hg38.fa.gz'))
